<?php

/*
  Cube object

  Copyright (C) 2007 Michiel van der Blonk - blonkm@gmail.com - All Rights Reserved
  Based on C++ code by Jo�l van Noort

  This software is licensed as per the MIT License below:
  MIT License: http://www.opensource.org/licenses/mit-license.php
  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
  Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * To generate the imagecube sticker codes, this program first simulates the
 * cube, using a  string that represents the cube state. Every elementary move
 * (UDLRFB) is then executed as five 4-cycles of stickers. Example: the
 * sticker permutations corresponding with the move U are:
 *
 *  U: (53, 51, 45, 47) (50, 52, 48, 46) (2, 38, 29, 11)
 *     (1, 37, 28, 10,) (0, 36, 27, 9)
 * 
 * The 5 cycles for one face are:
 * corner
 * edge
 * side left
 * side middle
 * side right
 */

// just for debugging
class Debug {

    var $mode;
    var $pre;
    var $level = 1;

    function out($s, $HTML = false, $insertNewlines = true, $pre = false) {
        if (isset($this) && !$pre) {
            $pre = $this->pre;
        }
        if (isset($this) && !$HTML) {
            $HTML = $this->mode;
        }

        if ($pre) {
            echo "<pre>";
        }
        if ($HTML) {
            echo '<div class="debug">';
            print_r($s);
            if ($insertNewlines) {
                echo '<br/>';
            } else {
                print_r($s);
            }
            if ($insertNewlines) {
                echo "\n";
            }
            echo '</div>';
        }
    }

    function testCube() {
        $c = new Cube();
        $c->performAlg('RULBRLUBRURLBU');
        echo $c->toImageCubeStickers();
    }

}

// post, get
class Page {

    function get($s, $defaultValue = '') {
        $value = $defaultValue;
        if (isset($_GET[$s])) {
            $value = $_GET[$s];
        }

        if (isset($_POST[$s])) {
            $value = $_POST[$s];
        }
        return $value;
    }

}

// handy functions
class Util {

    /** swaps two elements in an array
     * by reference
     */
    static function swap(&$var, $a, $b) {

        $temp = $var[$a];
        $var[$a] = $var[$b];
        $var[$b] = $temp;
    }

    /** rotates a string, e.g. cube -> ubec
     * 
     * @param type $s
     * @return type
     */
    function rotateString($s) {
        return substr($s, 1) . substr($s, 0, 1);
    }

    /** get the php version as either 3,4 or 5
     * 
     * @return type
     */
    function phpVersion() {
        return(int) substr(phpversion(), 0, 1);
    }

    function str_split($text, $split = 1) {
        // check precond
        if (!is_string($text)) {
            return false;
        }
        if (!is_numeric($split) && $split < 1) {
            return false;
        }

        $len = strlen($text);
        $ret = array();
        $s = 0;
        $e = $split;
        while ($s < $len) {
            $e = ($e < $len) ? $e : $len;
            // add next array element
            $ret[] = substr($text, $s, $e);
            $s += $e;
        }
        return $ret;
    }

    /** convert a string into an array of characters
     * not always will the s{x} format work well
     * so we need a 'real' array
     */
    static function toCharArray($str, $size = 1) {
        $chars = array();
        if (!function_exists("str_split")) {
            do {
                // get left part
                $chars[] = substr($str, 0, $size);
                // chop off and continue
                $str = substr($str, $size);
            } while ($str != "");
        } else {
            $chars = str_split($str, $size);
        }
        return $chars;
    }

    /** convert e.g. "UFL=U',UBR=U,UBL=U2,DFR=RUR'U',DFL=L U'L,DBR=BUB',DBL=B'U2B"
     * to an array with items UFL=>U', UBR=>U, etc..
     */
    function arraySplit($a, $delim = '=') {
        $ret = array();
        foreach ($a as $value) {
            $item = explode($delim, $value);
            $ret[$item[0]] = $item[1];
        }
        return $ret;
    }

}

// model of a Rubik's cube
class Cube {
    /* methods:
      Cube :                  constructor
      executePermutation :    4 cycle
      move:                   do 1 move, e.g. R, u, M, x
      clean:                  removes spaces, minus, quotes
      normalize:              makes R into R1 and R' into R3
      denormalize:              makes R1 into R' and R3 into R'
      inverse:                the algorithm reversed
      setMode:                define what stickers to show, e.g. only F2L
      execute:                execute a normalized alg (call move multiple times)
      toAppletStickers:       output the stickers for a RubikPlayer applet, based on standard layout (yellow on top)
      toImageCubeStickers:    output the stickers parameter for ImageCube (either Lars', Joel's or Michiel's)
      performAlg:             clean up and execute an algorithm string
      getSticker:             return the color of a sticker in a certain location
      stickersMatch:          check if two stickers given by location are the same color
      getPieceOrientations:   return all orientations of a piece. e.g. FUR => FUR,URF,RFU
      isPieceSolved:          check if a piece is in it's intended position
      getStickers:            return an array of stickers for a location, e.g. FUR => r,b,y
      locatePiece:            find a piece. E.g. the original RU piece is now in BD (RU=>BD)
     */

// default order of RubikPlayer faces is F, R, D, B, L, U
// we start with yellow on top
    var $defaultStickers = "rrrrrrrrrgggggggggwwwwwwwwwooooooooobbbbbbbbbyyyyyyyyy";
    var $emptyStickers = "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx";
    var $positions = "FLU,FU,FUR,FL,F,FR,FDL,FD,FRD,RFU,RU,RUB,RF,R,RB,RDF,RD,RBD,DLF,DF,DFR,DL,D,DR,DBL,DB,DRB,BRU,BU,BUL,BR,B,BL,BDR,BD,BLD,LBU,LU,LUF,LB,L,LF,LDB,LD,LFD,ULB,UB,UBR,UL,U,UR,UFL,UF,URF";
    var $edges = "UF,UR,UB,UL,FL,FR,BR,BL,DF,DR,DB,DL";
    var $edgesFlipped = "FU,RU,BU,LU,LF,RF,RB,LB,FD,RD,BD,LD";
    var $corners = "UFL,URF,UBR,ULB,DLF,DFR,DRB,DBL";
    var $stickers;
    var $mode;

    /* Algorithms for the solution using CFOP 
     * '*' means impossible configuration: previously solved edge is in this location
     * step 1: solve the cross
     */
    var $crossAlgs = "F2;,,,R'FR;U2,,,;U';L',,,D'LD;R,F2,,DR'D';R',B,R'UR;L;;D',*;D2,B2,X;D,L2,,*";
    var $crossAlgsFlipped = "U,,,L2F'L2;R',U2;U,U';L;F';F;R2,B,R2FR2;L',L,,;F;R,*;B,B',*;L'";
    // step 2: bring corner to UFR
    var $f2lAlgs1 = "UFL=U',UBR=U,UBL=U2,DFR=RUR'U',DFL=L U'L,DBR=BUB',DBL=B'U2B";
    var $f2lAlgs2 = "UF=;UL=R'U'R,UF'U'FU';UB=R'U2R;UR=B'RBR'U2;RF=U2RUR',RU2R'U';LF=R'L'ULR;RB=R'UR;LB=U2LU'L'";

    function Cube() {
        $this->stickers = $this->defaultStickers;
    }

    function reset() {
        $this->stickers = $this->defaultStickers;
    }
    
    /** this function uses swap function 3 times to make a 4-cycle of stickers:
     * 
     * @param type $a
     * @param type $b
     * @param type $c
     * @param type $d
     */
    function executePermutation($a, $b, $c, $d) {
        Util::swap($this->stickers, $a, $b);
        Util::swap($this->stickers, $a, $c);
        Util::swap($this->stickers, $a, $d);
    }

    /** perform a move and update stickers member variable
     * for multiple moves in the string $token the moves are done one by one
     * */
    function move($token) {
        $moveTransforms = array('u' => 'UEEE', 'd' => 'DE', 'f' => 'FS', 'b' => 'BSSS', 'l' => 'LM', 'r' => 'RMMM', 'x' => 'rLLL', 'y' => 'uDDD', 'z' => 'fBBB');

        // recursive calls to walk through the translated lowercase moves
        if (strlen($token) > 1) {
            foreach (Util::toCharArray($token) as $ch) {
                $this->move($ch);
            }
        }

        switch ($token) {
            case 'U':
                //one of the elementary move functions. Uses the executePermutation function
                //five times to make five 4-cycles of stickers:
                $this->executePermutation(53, 51, 45, 47);
                // top layer
                $this->executePermutation(50, 52, 48, 46);
                $this->executePermutation(2, 38, 29, 11);
                $this->executePermutation(1, 37, 28, 10);
                $this->executePermutation(0, 36, 27, 9);
                break;
            case 'D':
                $this->executePermutation(20, 26, 24, 18);
                $this->executePermutation(23, 25, 21, 19);
                $this->executePermutation(8, 17, 35, 44);
                $this->executePermutation(7, 16, 34, 43);
                $this->executePermutation(6, 15, 33, 42);
                break;
            case 'F':
                $this->executePermutation(2, 8, 6, 0);
                $this->executePermutation(5, 7, 3, 1);
                $this->executePermutation(9, 20, 44, 51);
                $this->executePermutation(12, 19, 41, 52);
                $this->executePermutation(15, 18, 38, 53);
                break;
            case 'B':
                $this->executePermutation(29, 35, 33, 27);
                $this->executePermutation(32, 34, 30, 28);
                $this->executePermutation(47, 36, 24, 17);
                $this->executePermutation(46, 39, 25, 14);
                $this->executePermutation(45, 42, 26, 11);
                break;
            case 'R':
                $this->executePermutation(11, 17, 15, 9);
                $this->executePermutation(14, 16, 12, 10);
                $this->executePermutation(2, 47, 33, 20);
                $this->executePermutation(5, 50, 30, 23);
                $this->executePermutation(8, 53, 27, 26);
                break;
            case 'L':
                $this->executePermutation(38, 44, 42, 36);
                $this->executePermutation(41, 43, 39, 37);
                $this->executePermutation(0, 18, 35, 45);
                $this->executePermutation(3, 21, 32, 48);
                $this->executePermutation(6, 24, 29, 51);
                break;
            case 'M':
                $this->executePermutation(4, 22, 31, 49);
                $this->executePermutation(1, 19, 34, 46);
                $this->executePermutation(7, 25, 28, 52);
                break;
            case 'E':
                $this->executePermutation(4, 13, 31, 40);
                $this->executePermutation(5, 14, 32, 41);
                $this->executePermutation(3, 12, 30, 39);
                break;
            case 'S':
                $this->executePermutation(49, 13, 22, 40);
                $this->executePermutation(50, 16, 21, 37);
                $this->executePermutation(48, 10, 23, 43);
                break;
            case '':
                break;
            default:
                // u,d,f,b,l,r,x,y,z
                $this->move($moveTransforms[$token]);
        }
    }

    /**
     * clean from certain tokens
     * note: only cleans up space, minus, quote
     * 
     * @param type $input
     * @return type
     */
    function clean($input) {
        if ($input == '') {
            return $input;
        }

        $tokens = " ()-\'";
        $input_clean = '';
        $prevCh = '';
        // delete magic quotes
        $input = str_replace("\\", '', $input);
        foreach (Util::toCharArray($input) as $ch) {
            if (strpos($tokens, $ch) === false) {
                $input_clean .= $ch;
            }

            // only add prime when right after a move token, e.g. R'
            // so skip R2' (becomes R2)
            if ($ch == '\'' && $prevCh != '2') {
                $input_clean .= $ch;
            }
            $prevCh = $ch;
        }
        return $input_clean;
    }

    /**
     * normalize algorithm
     * e.g. R2L'FD' becomes R2L3F1D3
     * 
     * @param type $input_clean
     * @return string
     */
    function normalize($input_clean) {
        $prevCh = "";
        $temp = str_replace('\'', '3', $input_clean);
        foreach (Util::toCharArray($temp) as $ch) {
            // when both ch and prev are moves, insert a 1 between them
            // RL becomes R1L
            if (preg_match("/\D/", $prevCh) && preg_match("/\D/", $ch)) {
                $alg_normal .= '1';
            }
            $alg_normal .= $ch;
            $prevCh = $ch;
        }
        if (preg_match("/\D/", $ch)) {
            $alg_normal .= '1';
        }
        return $alg_normal;
    }

    /**
     *
     * return a normalized algorithm to it's human readable state
     * e.g. R2L3F1D3 => R2L'FD'
     * 
     * @param type $input
     * @return type
     */
    function denormalize($input) {
        $algInverseReplaced = str_replace("3", "'", $input);
        $algSingleReplaced = str_replace("1", '', $algInverseReplaced);
        return $algSingleReplaced;
    }

    /**
     * reverse moves
     * used to show an algorithm that solves a cube
     * 
     * @param type $alg_normal
     * @return type
     */
    function inverse($alg_normal) {
        // split into separate moves
        $moves = str_split($alg_normal, 2);
        // reverse those
        $moves_inverse = array_reverse($moves);
        // back to string
        $alg_string = implode("", $moves_inverse);
        // inverse quantifiers
        $alg_hash = str_replace('1', '#', $alg_string);
        // temp ch for 3
        $alg_temp_3 = str_replace('3', '1', $alg_hash);
        $alg_inverse = str_replace('#', '3', $alg_temp_3);

        return $alg_inverse;
    }

    /**
     * several predefined sticker modes
     * @param type $mode
     */
    function setMode($mode = '') {
        if ($mode != '') {
            $this->mode = $mode;
        }

        switch (strtoupper($this->mode)) {
            case "CROSS":
                $this->stickers = "xxxxrxxrxxxxxgxxgxxwxwwwxwxxxxxoxxoxxxxxbxxbxxxxxyxxxx";
                break;
            case "FL": // first layer
                $this->stickers = "xxxxxxrrrxxxxxxgggwwwwwwwwwxxxxxxoooxxxxxxbbbxxxxxxxxx";
                break;
            case "F2L": // first two layers
                $this->stickers = "xxxrrrrrrxxxggggggwwwwwwwwwxxxooooooxxxbbbbbbxxxxyxxxx";
                break;
            case "SHORTCUT": // don't remember what this one is for
                $this->stickers = "xxxxrrxrrxxxggxggxxwwwwwxwxxxxxoxxoxxxxxbxxbxxxxxyxxxx";
                break;
            case "OLL": // orientation of last layer
                $this->stickers = "xxxrrrrrrxxxggggggwwwwwwwwwxxxooooooxxxbbbbbbyyyyyyyyy";
                break;
            case "PLL": // permutation of last layer
                // $this->stickers = "rrrrrrrrrgggggggggwwwwwwwwwooooooooobbbbbbbbbyyyyyyyyy";
                $this->stickers = "rrrxxxxxxgggxxxxxxxxxxxxxxxoooxxxxxxbbbxxxxxxyyyyyyyyy";
                break;
        }
    }

    /**
     * execute an algorithm
     * assumes the $alg is normalized
     * 
     * @param type $alg
     * @param type $mode
     * @return type
     */
    function execute($alg, $mode = '') {
        // prepare with a certain mode if given
        if ($mode != '') {
            $this->setMode($mode);
        }

        // walk through all moves in a normalized alg
        for ($i = 0; $i < strlen($alg); $i+=2) {
            // define operation (move) and quantifier
            $op = $alg[$i];
            $q = $alg[$i + 1];
            // move a face n times
            foreach (range(1, $q) as $value)
                $this->move($op);
        }
        return $this->stickers;
    }

// 
// 
// 

    /**
     * applet stickers for RubikPlayer are
     * the same as the stickers member variable
     * caveat: this does not work for new applet Rubikplayer 7
     * 
     * @param type $delta
     * @param type $separator
     * @param type $outputAsPositions
     * @return type
     */
    function toAppletStickers($delta = false, $separator = '', $outputAsPositions = false) {
        if ($delta) {
            // first convert all strings to arrays
            // in order to take advantage of powerful array functions
            $aDefault = Util::toCharArray($this->defaultStickers);
            $aCurrent = Util::toCharArray($this->stickers);
            $aEmpty = Util::toCharArray($this->emptyStickers);
            // get the stickers that have changed
            $aDelta = array_diff_assoc($aDefault, $aCurrent);
            if ($outputAsPositions) {
                $aPositions = explode(",", $this->positions);
                $changedStickers = array_intersect_key($aPositions, $aDelta);
                $out = array_combine($changedStickers, $aDelta);
            }
            // add all other stickers as 'x' (empty sticker)
            $aCombined = $aDelta + $aEmpty;
            // keys have changed, re-sort in natural order
            ksort($aCombined);
            // convert back to string
            $ret = join($separator, $aCombined);
            return $ret;
        } else {
            return join($separator, Util::toCharArray($this->stickers));
        }
    }

    /**
     * return only visible faces for imagecube
     * i.e. U, F, R
     * 
     * @return string
     */
    function toImageCubeStickers() {
        // split into 9 faces
        $faces = Util::toCharArray($this->stickers, 9);
        // create string for up, right, front
        $upFace = $faces[5];
        $frontFace = $faces[0];
        $rightFace = $faces[1];
        $imageStickers = $upFace . $frontFace . $rightFace;

        return $imageStickers;
    }

    /**
     * return sticker color by index (e.g. 3) or by sticker position (e.g. FDL)
     * 
     * @param type $position
     * @return type
     */
    function getSticker($position) {
        if (is_numeric($position)) {
            return $this->stickers[(int) $position];
        } else {
            $aPositions = explode(",", $this->positions);
            $key = array_search($position, $aPositions);
            if ($key !== false) {
                return $this->stickers[$key];
            } else {
                trigger_error("no such position: " . $position, E_USER_ERROR);
            }
        }
    }

    /**
     * check whether two stickers match color
     * @param type $s1
     * @param type $s2
     * @return type
     */
    function stickersMatch($s1, $s2) {
        return $this->getSticker($s1) == $this->getSticker($s2);
    }

    /**
     * get all possible orientations of a piece
     * e.g. RFU => RFU,FUR,URF
     * 
     * @param type $piece
     * @return type
     */
    function getPieceOrientations($piece) {
        $ret = array();
        $rotatedPiece = $piece;
        do {
            $ret[] = $rotatedPiece;
            $rotatedPiece = Util::rotateString($rotatedPiece);
        } while ($rotatedPiece != $piece);
        return $ret;
    }

    /**
     * check whether a piece is in it's intended position
     *  i.e. match it with the center pieces' colors
     * $piece is in format of position, e.g. RFU
     * 
     * @param type $piece
     * @return boolean
     */
    function isPieceSolved($piece) {
        $rotatedPiece = $piece;
        do {
            if (!$this->stickersMatch($piece, $piece[0])) {
                return false;
            }
            $rotatedPiece = Util::rotateString($rotatedPiece);
        } while ($rotatedPiece != $piece);
        return true;
    }

    /**
     * return an array of stickers in a certain position
     * two or three colors (for edge or corner)
     * matching the colors of a piece
     * e.g. RFU => g,r,y.
     * 
     * @param type $position
     * @return type
     */
    function getStickers($position) {
        $ret = array();
        $pieceOrientations = $this->getPieceOrientations($position);
        foreach ($pieceOrientations as $sticker) {
            $ret[] = $this->getSticker($sticker);
        }
        return $ret;
    }

    /**
     * find a piece on the cube
     * e.g. RU edge is located in FL
     * 
     * @param type $piece
     * @return type
     */
    function locatePiece($piece) {
        $ret = false;
        $aPositions = explode(",", $this->positions);
        // get stickers on default cube,
        // these are the target stickers we look for
        $newCube = new Cube;
        $pieceStickers = $newCube->getStickers($piece);
        //go through all pieces
        foreach ($aPositions as $p) {
            $positionStickers = $this->getStickers($p);
            //if colors match
            if ($pieceStickers == $positionStickers) {
                $ret = $p;
            }
        }
        return $ret;
    }

    /**
     * 
     * perform an algorithm given
     * done in the steps
     * 1. clean up
     * 2. normalize (add quantifiers)
     * 3. optionally inverse
     * 4. execute
     * 
     * @param type $input
     * @param type $inverse
     */
    function performAlg($input, $inverse = false) {
        $input_clean = "";
        $alg_normal = "";
        $alg_inverse = "";

        // clean input leaves only the moves
        $alg = $this->clean($input);

        // process to get normalized alg (e.g. R2F1L3)
        $alg = $this->normalize($alg);

        // process inverse of alg
        if ($inverse) {
            $alg = $this->inverse($alg);
        }

        // perform inverse on the stickers
        $this->stickers = $this->execute($alg);
    }

    /**
     * compact an alg
     * e.g. RRR => R3, and L2L3 => L1
     * 
     * @param type $input
     * @return string
     */
    function compact($input) {
        $input_clean = "";
        $alg_normal = "";
        $alg_inverse = "";

        if ($input == '') {
            return '';
        }

        // clean input leaves only the moves
        $alg = $this->clean($input);

        // process to get normalized alg (e.g. R2F1L3)
        $alg = $this->normalize($alg);

        // create array of operations
        $ops = array();
        for ($i = 0; $i < strlen($alg); $i+=2) {
            $ops[$i / 2]['move'] = $alg[$i];
            $ops[$i / 2]['count'] = $alg[$i + 1];
        }

        $prevKey='';
        $prevOp=array();
        // merge operations to remove duplicates
        // e.g. D2D2 => D4 => D0
        foreach ($ops as $key => $op) {
            // if a duplicate is found
            if ($op['move'] == $prevOp['move']) {
                // add the moves
                $ops[$key]['count'] += $prevOp['count'];
                // mod 4 to remove 360 turns
                $ops[$key]['count'] %= 4;
                // remove the previous one, it is now
                // included in the current operation
                unset($ops[$prevKey]);
            }
            $prevOp = $op;
            $prevKey = $key;
        }

        // make into string again
        $out = '';
        foreach ($ops as $op) {
            $out .= $op['move'] . $op['count'];
        }

        return $out;
    }

}

?>